import numpy as np


def np_ndvi(red: np.ndarray, nir: np.ndarray) -> np.ndarray:
    return (nir - red) / (nir + red)


def np_gndvi(green: np.ndarray, nir: np.ndarray) -> np.ndarray:
    return (nir - green) / (nir + green)


def np_ndwi(nir: np.ndarray, swir1: np.ndarray) -> np.ndarray:
    return (nir - swir1) / (nir + swir1)


def np_savi(red: np.ndarray, nir: np.ndarray, L: float = 0.5) -> np.ndarray:
    return ((nir - red) / (nir + red + L)) * (1 + L)


def np_evi(blue: np.ndarray, red: np.ndarray, nir: np.ndarray) -> np.ndarray:
    return 2.5 * (nir - red) / (nir + 6 * red - 7.5 * blue + 1)


def np_evi2(red: np.ndarray, nir: np.ndarray) -> np.ndarray:
    return 2.5 * (nir - red) / (nir + 2.4 * red + 1)


def np_msavi(red: np.ndarray, nir: np.ndarray) -> np.ndarray:
    return (2 * nir + 1 - np.sqrt((2 * nir + 1) ** 2 - 8 * (nir - red))) / 2


def np_ndre(red_edge: np.ndarray, nir: np.ndarray) -> np.ndarray:
    return (nir - red_edge) / (nir + red_edge)


def np_mcari(green: np.ndarray, red: np.ndarray, nir: np.ndarray) -> np.ndarray:
    return ((nir - red) - 0.2 * (nir - green)) * (nir / red)


def np_gci(green: np.ndarray, nir: np.ndarray) -> np.ndarray:
    return (nir / green) - 1


def np_bsi(blue: np.ndarray, red: np.ndarray, nir: np.ndarray, swir1: np.ndarray) -> np.ndarray:
    return ((swir1 + red) - (nir + blue)) / ((swir1 + red) + (nir + blue))


def np_ci_red(red: np.ndarray, nir: np.ndarray) -> np.ndarray:
    return (nir / red) - 1


def np_ci_green(green: np.ndarray, nir: np.ndarray) -> np.ndarray:
    return (nir / green) - 1


def np_osavi(red: np.ndarray, nir: np.ndarray) -> np.ndarray:
    return (nir - red) / (nir + red + 0.16)


def np_arvi(blue: np.ndarray, red: np.ndarray, nir: np.ndarray) -> np.ndarray:
    return (nir - (2 * red - blue)) / (nir + (2 * red - blue))


def np_vhvv(vv: np.ndarray, vh: np.ndarray) -> np.ndarray:
    return vh / vv


ALL_NUMPY_INDICES = {
    "ndvi": np_ndvi,
    "gndvi": np_gndvi,
    "ndwi": np_ndwi,
    "savi": np_savi,
    "evi": np_evi,
    "evi2": np_evi2,
    "msavi": np_msavi,
    "ndre": np_ndre,
    "mcari": np_mcari,
    "gci": np_gci,
    "bsi": np_bsi,
    "ci_red": np_ci_red,
    "ci_green": np_ci_green,
    "osavi": np_osavi,
    "arvi": np_arvi,
    "vhvv": np_vhvv,
}



import concurrent.futures
import logging
import logging.handlers
import queue
from functools import partial

import ee
import numpy as np
import pandas as pd
from shapely import Polygon
from tqdm.std import tqdm

from agrigee_lite.ee_utils import ee_img_to_numpy
from agrigee_lite.misc import cached
from agrigee_lite.sat.abstract_satellite import AbstractSatellite


# @cached
def download_multiple_images(
    geometry: Polygon,
    start_date: pd.Timestamp | str,
    end_date: pd.Timestamp | str,
    satellite: AbstractSatellite,
    invalid_images_threshold: float = 0.5,
    num_threads_rush: int = 30,
    num_threads_retry: int = 10,
) -> np.ndarray:
    log_queue: queue.Queue[logging.LogRecord] = queue.Queue(-1)

    file_handler = logging.FileHandler("logging.log", mode="a")
    formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
    file_handler.setFormatter(formatter)

    queue_listener = logging.handlers.QueueListener(log_queue, file_handler)
    queue_listener.start()

    queue_handler = logging.handlers.QueueHandler(log_queue)
    logger = logging.getLogger("logger_sits")
    logger.setLevel(logging.ERROR)
    logger.addHandler(queue_handler)
    logger.propagate = False

    ee_geometry = ee.Geometry(geometry.__geo_interface__)
    ee_feature = ee.Feature(
        ee_geometry,
        {"s": start_date, "e": end_date, "0": 1},
    )
    ee_expression = satellite.imageCollection(ee_feature)

    max_valid_pixels = ee_expression.aggregate_max("ZZ_USER_VALID_PIXELS")
    threshold = ee.Number(max_valid_pixels).multiply(invalid_images_threshold)
    ee_expression = ee_expression.filter(ee.Filter.gte("ZZ_USER_VALID_PIXELS", threshold))

    image_names = ee_expression.aggregate_array("ZZ_USER_TIME_DUMMY").getInfo()

    image_indexes = [
        (n, image_index)
        for n, image_index in enumerate(ee.data.computeValue(ee_expression.aggregate_array("system:index")))
    ]
    image_indexes_with_errors = []
    all_images = [np.array([]) for _ in range(len(image_indexes))]

    def process_download(i: int, image_index: str) -> tuple[np.ndarray, int, str]:
        try:
            img = ee_img_to_numpy(
                ee_expression.filter(ee.Filter.eq("system:index", image_index)).first().clip(ee_geometry),
                ee_geometry,
                10,
            )
            return img, i, image_index  # noqa: TRY300
        except Exception as e:
            logger.error(f"download_multiple_images_multithread_{i}_{satellite.shortName} = {e}")  # noqa: TRY400
            return np.array([]), i, image_index

    def run_downloads(image_indexes: list[tuple[int, str]], num_threads: int, desc: str) -> None:
        nonlocal all_images
        error_count = 0

        with concurrent.futures.ThreadPoolExecutor(num_threads) as executor:
            futures = [executor.submit(partial(process_download, n, image_index)) for n, image_index in image_indexes]

            pbar = tqdm(concurrent.futures.as_completed(futures), total=len(futures), desc=desc)
            for future in pbar:
                result_img, i, image_index = future.result()
                if result_img.shape[0] == 0:
                    error_count += 1
                    image_indexes_with_errors.append((i, image_index))
                    pbar.set_postfix({"errors": error_count})
                else:
                    all_images[i] = result_img

    run_downloads(image_indexes, num_threads=num_threads_rush, desc="Downloading")

    if image_indexes_with_errors:
        run_downloads(
            image_indexes_with_errors,
            num_threads=num_threads_retry,
            desc="Re-running failed downloads",
        )

    return np.stack(all_images), image_names



import geopandas as gpd
import numpy as np
import pandas as pd

import agrigee_lite as agl
from agrigee_lite.sat.abstract_satellite import AbstractSatellite


def assert_np_array_equivalence(arr1: np.ndarray, arr2: np.ndarray, threshold: float = 0.1) -> None:
    arr1 = arr1.flatten()
    arr2 = arr2.flatten()

    if arr1.shape != arr2.shape:
        raise AssertionError(f"Shape mismatch: {arr1.shape} vs {arr2.shape}")

    bound1 = arr2 * (1 - threshold)
    bound2 = arr2 * (1 + threshold)
    lower_bound = np.minimum(bound1, bound2)
    upper_bound = np.maximum(bound1, bound2)

    valid = (arr1 >= lower_bound) & (arr1 <= upper_bound)

    if not np.all(valid):
        invalid_indices = np.where(~valid)[0]
        error_pct = 100 * len(invalid_indices) / len(arr1)

        msg_lines = [
            f"Arrays differ: {len(invalid_indices)} of {len(arr1)} values outside threshold ({error_pct:.2f}%)",
        ]

        for i in invalid_indices[:5]:
            msg_lines.append(
                f"[{i}] Downloaded={arr1[i]:.4f}, Original={arr2[i]:.4f}, "
                f"Allowed=[{lower_bound[i]:.4f}, {upper_bound[i]:.4f}]"
            )

        raise AssertionError("\n".join(msg_lines))


def assert_df_equivalence(
    df1: pd.DataFrame | gpd.GeoDataFrame, df2: pd.DataFrame | gpd.GeoDataFrame, threshold: float = 0.1
) -> None:
    cols1 = set(df1.columns)
    cols2 = set(df2.columns)

    missing_in_df2 = cols1 - cols2
    extra_in_df2 = cols2 - cols1

    if missing_in_df2 or extra_in_df2:
        msg = ["Column mismatch detected:"]
        if missing_in_df2:
            msg.append(f"- Missing in second DataFrame: {sorted(missing_in_df2)}")
        if extra_in_df2:
            msg.append(f"- Additional in second DataFrame: {sorted(extra_in_df2)}")
        raise AssertionError("\n".join(msg))

    arr1 = df1[sorted(cols1)].to_numpy()
    arr2 = df2[sorted(cols1)].to_numpy()

    assert_np_array_equivalence(arr1, arr2, threshold)


def get_all_satellites_for_test() -> list[AbstractSatellite]:
    return [
        agl.sat.Sentinel2(),
        agl.sat.Sentinel2(use_sr=True),
        agl.sat.Sentinel1(),
        agl.sat.Sentinel1(ascending=False),
    ]


def get_all_date_types_for_test() -> list[str]:
    return ["doy", "year", "fyear"]


def get_all_reducers_for_test() -> list[str]:
    return ["min", "max", "mean", "median", "mode", "std", "var", "p2", "p98", "kurt", "skew"]



import numpy as np
import pandas as pd
from tqdm.std import tqdm


class AbstractMultiModalDataset:
    def __init__(self, long_sits: pd.DataFrame, original_gdf_lenght: int):
        grouped = long_sits.groupby("indexnum")

        max_seq_len = grouped.size().max()
        num_bands = len(long_sits.columns) - 2

        self.X = np.zeros((original_gdf_lenght, max_seq_len, num_bands), dtype=np.float16)
        self.T = np.zeros((original_gdf_lenght, max_seq_len), dtype="datetime64[D]")

        for idx, group in tqdm(grouped):
            group_sorted = group.sort_values("timestamp")
            seq_len = len(group_sorted)

            self.X[idx, :seq_len, :] = group_sorted.drop(columns=["timestamp", "indexnum"]).to_numpy(dtype=np.float16)
            self.T[idx, :seq_len] = group_sorted["timestamp"].to_numpy().astype("datetime64[D]")

    def __len__(self):
        raise NotImplementedError("Subclasses should implement this method.")

    def __getitem__(self, idx):
        raise NotImplementedError("Subclasses should implement this method.")



import ee
import pytest


@pytest.fixture(scope="session", autouse=True)
def test_ee_authenticate() -> None:
    ee.Initialize(opt_url="https://earthengine-highvolume.googleapis.com", project="ee-paulagibrim")
    assert ee.data._initialized



import ee

from agrigee_lite.vegetation_indices import VEGETATION_INDICES


class AbstractSatellite:
    def __init__(self) -> None:
        self.startDate = ""
        self.endDate = ""
        self.shortName = "IDoNotExist"
        self.availableBands: dict[str, str] = {}
        self.selectedBands: list[tuple[str, str, str]] = {}
        self.imageCollectionName = ""
        self.pixelSize: int = 0

    def imageCollection(self, ee_feature: ee.Feature) -> ee.ImageCollection:
        return ee.ImageCollection()

    def compute(
        self,
        ee_feature: ee.Feature,
        subsampling_max_pixels: float,
        reducers: list[str] | None = None,
    ) -> ee.FeatureCollection:
        return ee.FeatureCollection()

    def log_dict(self) -> dict:
        return {self.__class__.__name__: self.__dict__}

    @property
    def availableIndices(self) -> dict[str, str]:
        return {
            name: idx["expression"]
            for name, idx in VEGETATION_INDICES.items()
            if idx["required_bands"].issubset(self.availableBands.keys())
        }

    def __str__(self) -> str:
        return self.shortName

    def __repr__(self) -> str:
        return self.shortName


class OpticalSatellite(AbstractSatellite):
    def __init__(self) -> None:
        super().__init__()
        self.dateType = "optical"


class RadarSatellite(AbstractSatellite):
    def __init__(self) -> None:
        super().__init__()
        self.dateType = "radar"


class DataSourceSatellite(AbstractSatellite):
    def __init__(self) -> None:
        super().__init__()
        self.dateType = "dataSource"



from functools import partial

import ee

from agrigee_lite.ee_utils import (
    ee_add_indexes_to_image,
    ee_cloud_probability_mask,
    ee_filter_img_collection_invalid_pixels,
    ee_get_number_of_pixels,
    ee_get_reducers,
    ee_map_bands_and_doy,
    ee_safe_remove_borders,
)
from agrigee_lite.sat.abstract_satellite import OpticalSatellite


class Sentinel2(OpticalSatellite):
    """
    Satellite abstraction for Sentinel-2 (HARMONIZED collections).

    Sentinel-2 is a constellation of twins Earth observation satellites,
    operated by ESA, designed for land monitoring, vegetation, soil, water cover, and coastal areas.

    Parameters
    ----------
    bands : list of str, optional
        List of bands to select. Defaults to all 10 bands most used for vegetation and soil analysis.
    use_sr : bool, defaults to False
        If True, uses surface reflectance (BOA, 'S2_SR_HARMONIZED').
        If False, uses top-of-atmosphere reflectance ('S2_HARMONIZED').

    Satellite Information
    ---------------------
    +-----------------------------------+------------------------+
    | Field                             | Value                  |
    +-----------------------------------+------------------------+
    | Name                              | Sentinel-2             |
    | Revisit Time                      | 5 days                 |
    | Revisit Time (cloud-free estimate) | ~7 days               |
    | Pixel Size                        | 10 meters              |
    | Coverage                          | Global                 |
    +-----------------------------------+------------------------+

    Collection Dates
    ----------------
    +------------------+------------+----------------+
    | Collection Type  | Start Date | End Date        |
    +------------------+------------+----------------+
    | TOA (Top of Atmosphere) | 2016-01-01 | present    |
    | SR (Surface Reflectance) | 2019-01-01 | present   |
    +------------------+------------+----------------+

    Band Information
    ------------
    +------------+---------------+--------------+----------------------+
    | Band Name  | Original Band | Resolution   | Spectral Wavelength   |
    +------------+---------------+--------------+----------------------+
    | blue       | B2            | 10 m        | 492 nm                |
    | green      | B3            | 10 m        | 559 nm                |
    | red        | B4            | 10 m        | 665 nm                |
    | re1        | B5            | 20 m        | 704 nm                |
    | re2        | B6            | 20 m        | 739 nm                |
    | re3        | B7            | 20 m        | 780 nm                |
    | nir        | B8            | 10 m        | 833 nm                |
    | re4        | B8A           | 20 m        | 864 nm                |
    | swir1      | B11           | 20 m        | 1610 nm               |
    | swir2      | B12           | 20 m        | 2186 nm               |
    +------------+---------------+--------------+----------------------+
    """

    def __init__(
        self,
        bands: list[str] | None = None,
        indices: list[str] | None = None,
        use_sr: bool = True,
    ):
        if bands is None:
            bands = [
                "blue",
                "green",
                "red",
                "re1",
                "re2",
                "re3",
                "nir",
                "re4",
                "swir1",
                "swir2",
            ]

        if indices is None:
            indices = []

        super().__init__()
        self.useSr = use_sr
        self.imageCollectionName = "COPERNICUS/S2_SR_HARMONIZED" if use_sr else "COPERNICUS/S2_HARMONIZED"
        self.pixelSize: int = 10

        self.startDate: str = "2019-01-01" if use_sr else "2016-01-01"
        self.endDate: str = "2050-01-01"
        self.shortName: str = "s2sr" if use_sr else "s2"

        self.availableBands: dict[str, str] = {
            "blue": "B2",
            "green": "B3",
            "red": "B4",
            "re1": "B5",
            "re2": "B6",
            "re3": "B7",
            "nir": "B8",
            "re4": "B8A",
            "swir1": "B11",
            "swir2": "B12",
        }

        self.selectedBands: list[tuple[str, str, str]] = [
            (band, f"{(n + 10):02}_{band}") for n, band in enumerate(bands)
        ]

        self.selectedIndices: list[str] = [
            (self.availableIndices[indice_name], indice_name, f"{(n + 40):02}_{indice_name}")
            for n, indice_name in enumerate(indices)
        ]

    def imageCollection(self, ee_feature: ee.Feature) -> ee.ImageCollection:
        ee_geometry = ee_feature.geometry()

        ee_start_date = ee_feature.get("s")
        ee_end_date = ee_feature.get("e")

        ee_filter = ee.Filter.And(ee.Filter.bounds(ee_geometry), ee.Filter.date(ee_start_date, ee_end_date))

        s2_img = (
            ee.ImageCollection(self.imageCollectionName)
            .filter(ee_filter)
            .select(
                list(self.availableBands.values()),
                list(self.availableBands.keys()),
            )
        )

        s2_img = s2_img.map(lambda img: ee.Image(img).addBands(ee.Image(img).divide(10000), overwrite=True))

        if self.selectedIndices:
            s2_img = s2_img.map(
                partial(ee_add_indexes_to_image, indexes=[expression for (expression, _, _) in self.selectedIndices])
            )

        s2_img = s2_img.select(
            [natural_band_name for natural_band_name, _ in self.selectedBands]
            + [indice_name for _, indice_name, _ in self.selectedIndices],
            [numeral_band_name for _, numeral_band_name in self.selectedBands]
            + [numeral_indice_name for _, _, numeral_indice_name in self.selectedIndices],
        )

        s2_cloud_mask = (
            ee.ImageCollection("GOOGLE/CLOUD_SCORE_PLUS/V1/S2_HARMONIZED")
            .filter(ee_filter)
            .select(["cs_cdf"], ["cloud"])
        )

        s2_img = s2_img.combine(s2_cloud_mask)

        s2_img = s2_img.map(lambda img: ee_cloud_probability_mask(img, 0.7, True))
        s2_img = ee_filter_img_collection_invalid_pixels(s2_img, ee_geometry, self.pixelSize, 20)

        return ee.ImageCollection(s2_img)

    def compute(
        self,
        ee_feature: ee.Feature,
        subsampling_max_pixels: float,
        reducers: list[str] | None = None,
    ) -> ee.FeatureCollection:
        ee_geometry = ee_feature.geometry()
        ee_geometry = ee_safe_remove_borders(ee_geometry, self.pixelSize, 35000)
        ee_feature = ee_feature.setGeometry(ee_geometry)

        s2_img = self.imageCollection(ee_feature)

        features = s2_img.map(
            partial(
                ee_map_bands_and_doy,
                ee_feature=ee_feature,
                pixel_size=self.pixelSize,
                subsampling_max_pixels=ee_get_number_of_pixels(ee_geometry, subsampling_max_pixels, self.pixelSize),
                reducer=ee_get_reducers(reducers),
            )
        )

        return features



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from shapely import Polygon

from agrigee_lite.get.image import download_multiple_images
from agrigee_lite.sat.abstract_satellite import OpticalSatellite


def visualize_multiple_images(
    geometry: Polygon,
    start_date: pd.Timestamp | str,
    end_date: pd.Timestamp | str,
    satellite: OpticalSatellite,
    invalid_images_threshold: float = 0.5,
    contrast: float = 1.3,
    num_threads_rush: int = 30,
    num_threads_retry: int = 10,
) -> None:
    images, image_names = download_multiple_images(
        geometry, start_date, end_date, satellite, invalid_images_threshold, num_threads_rush, num_threads_retry
    )

    images = np.clip(images * contrast, 0, 1)

    images_per_row = 10
    num_rows = (len(image_names) // images_per_row) + (len(image_names) % images_per_row > 0)

    fig, ax = plt.subplots(
        num_rows, images_per_row, figsize=(25, (len(image_names) // images_per_row) * 5), sharex=True, sharey=True
    )
    for i in range(num_rows * images_per_row):
        if i < len(image_names):
            name = image_names[i]
            image = images[i]

            ax[i // images_per_row, i % images_per_row].imshow(image)
            ax[i // images_per_row, i % images_per_row].set_title(name)
            ax[i // images_per_row, i % images_per_row].axis("off")
        else:
            fig.delaxes(ax[i // images_per_row, i % images_per_row])

    plt.tight_layout()
    plt.show()



import concurrent.futures
import getpass
import json
import logging
import logging.handlers
import pathlib
import queue
from functools import partial

import ee
import geopandas as gpd
import numpy as np
import pandas as pd
import pandera as pa
from shapely import Polygon
from smart_open import open  # noqa: A004
from tqdm.std import tqdm

from agrigee_lite.ee_utils import ee_gdf_to_feature_collection, ee_get_tasks_status
from agrigee_lite.misc import (
    add_indexnum_column,
    create_gdf_hash,
    log_dict_function_call_summary,
    quadtree_clustering,
    remove_underscore_in_df,
)
from agrigee_lite.sat.abstract_satellite import AbstractSatellite
from agrigee_lite.task_manager import GEETaskManager


# @cached # Doesn't work with lists as parameters :(
def download_single_sits(
    geometry: Polygon,
    start_date: pd.Timestamp | str,
    end_date: pd.Timestamp | str,
    satellite: AbstractSatellite,
    reducers: list[str] | None = None,
    subsampling_max_pixels: float = 1_000,
) -> pd.DataFrame:
    start_date = start_date.strftime("%Y-%m-%d") if isinstance(start_date, pd.Timestamp) else start_date
    end_date = end_date.strftime("%Y-%m-%d") if isinstance(end_date, pd.Timestamp) else end_date

    if start_date < satellite.startDate or end_date > satellite.endDate:
        raise ValueError(  # noqa: TRY003
            f"Start date {start_date} or end date {end_date} is outside the satellite's range "
            f"({satellite.startDate} to {satellite.endDate})"
        )

    ee_feature = ee.Feature(
        geometry.__geo_interface__,
        {"s": start_date, "e": end_date, "0": 0},
    )
    ee_expression = satellite.compute(ee_feature, reducers=reducers, subsampling_max_pixels=subsampling_max_pixels)

    sits_df = ee.data.computeFeatures({"expression": ee_expression, "fileFormat": "PANDAS_DATAFRAME"}).drop(
        columns=["geo"]
    )

    remove_underscore_in_df(sits_df)

    return sits_df


def download_multiple_sits(
    gdf: gpd.GeoDataFrame,
    satellite: AbstractSatellite,
    reducers: list[str] | None = None,
    subsampling_max_pixels: float = 1_000,
) -> pd.DataFrame:
    add_indexnum_column(gdf)

    fc = ee_gdf_to_feature_collection(gdf)
    ee_expression = ee.FeatureCollection(
        fc.map(
            partial(
                satellite.compute,
                reducers=reducers,
                subsampling_max_pixels=subsampling_max_pixels,
            )
        )
    ).flatten()
    sits_df = ee.data.computeFeatures({"expression": ee_expression, "fileFormat": "PANDAS_DATAFRAME"}).drop(
        columns=["geo"]
    )

    remove_underscore_in_df(sits_df)

    return sits_df


def download_multiple_sits_multithread(
    gdf: gpd.GeoDataFrame,
    satellite: AbstractSatellite,
    reducers: list[str] | None = None,
    subsampling_max_pixels: float = 1_000,
    mini_chunksize: int = 10,
    num_threads_rush: int = 30,
    num_threads_retry: int = 10,
    pbar: tqdm | None = None,
) -> pd.DataFrame:
    add_indexnum_column(gdf)
    log_queue: queue.Queue[logging.LogRecord] = queue.Queue(-1)

    file_handler = logging.FileHandler("logging.log", mode="a")
    formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
    file_handler.setFormatter(formatter)

    queue_listener = logging.handlers.QueueListener(log_queue, file_handler)
    queue_listener.start()

    queue_handler = logging.handlers.QueueHandler(log_queue)
    logger = logging.getLogger("logger_sits")
    logger.setLevel(logging.ERROR)
    logger.addHandler(queue_handler)
    logger.propagate = False

    indexes_with_errors: list[int] = []
    whole_result_df = pd.DataFrame()
    num_chunks = (len(gdf) + mini_chunksize - 1) // mini_chunksize
    all_chunk_ids = list(range(num_chunks))

    def process_download(gdf_chunk: gpd.GeoDataFrame, i: int) -> tuple[pd.DataFrame, int]:
        try:
            result_chunk = download_multiple_sits(
                gdf_chunk,
                satellite,
                reducers=reducers,
                subsampling_max_pixels=subsampling_max_pixels,
            )
            return result_chunk, i  # noqa: TRY300
        except KeyboardInterrupt:
            raise KeyboardInterrupt  # noqa: B904
        except Exception as e:
            logger.error(f"download_multiple_sits_multithread_{i}_{satellite.shortName} = {e}")  # noqa: TRY400
            return pd.DataFrame(), i

    def run_downloads(chunk_ids: list[int], num_threads: int) -> None:
        nonlocal whole_result_df
        error_count = 0

        with concurrent.futures.ThreadPoolExecutor(num_threads) as executor:
            futures = [
                executor.submit(
                    partial(
                        process_download,
                        gdf.iloc[i * mini_chunksize : (i + 1) * mini_chunksize].reset_index(drop=True),
                        i,
                    )
                )
                for i in chunk_ids
            ]

            for future in concurrent.futures.as_completed(futures):
                result_df_chunk, i = future.result()
                if result_df_chunk.empty:
                    error_count += 1
                    indexes_with_errors.append(i)

                    if pbar is not None:
                        pbar.set_postfix({"errors": error_count})
                else:
                    whole_result_df = pd.concat([whole_result_df, result_df_chunk])

                    if pbar is not None:
                        pbar.update(result_df_chunk.indexnum.nunique())

    run_downloads(all_chunk_ids, num_threads=num_threads_rush)

    if indexes_with_errors:
        run_downloads(indexes_with_errors, num_threads=num_threads_retry)

    return whole_result_df


# Dead code, but kept in case we need it in the future
async def __download_multiple_sits_async(
    gdf: gpd.GeoDataFrame,
    satellite: AbstractSatellite,
    reducers: list[str] | None = None,
    subsampling_max_pixels: float = 1_000,
    mini_chunksize: int = 10,
    initial_concurrency: int = 30,
    retry_concurrency: int = 10,
    initial_timeout: float = 20,
    retry_timeout: float = 10,
) -> pd.DataFrame:
    import anyio

    add_indexnum_column(gdf)
    log_queue: queue.Queue[logging.LogRecord] = queue.Queue(-1)

    file_handler = logging.FileHandler("logging.log", mode="a")
    formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
    file_handler.setFormatter(formatter)

    queue_listener = logging.handlers.QueueListener(log_queue, file_handler)
    queue_listener.start()

    queue_handler = logging.handlers.QueueHandler(log_queue)
    logger = logging.getLogger("logger_sits")
    logger.setLevel(logging.ERROR)
    logger.addHandler(queue_handler)
    logger.propagate = False

    num_chunks = (len(gdf) + mini_chunksize - 1) // mini_chunksize
    all_chunk_ids = list(range(num_chunks))
    whole_result_df = pd.DataFrame()
    indexes_with_errors: list[int] = []

    async def run_download(chunk_ids: list[int], concurrency: int, timeout: float) -> None:
        sem = anyio.Semaphore(concurrency)

        async def download_task(chunk_id: int) -> None:
            nonlocal whole_result_df

            async with sem:
                start_idx = chunk_id * mini_chunksize
                end_idx = min(start_idx + mini_chunksize, len(gdf))

                gdf_chunk = gdf.iloc[start_idx:end_idx].reset_index(drop=True)

                try:
                    with anyio.fail_after(timeout):
                        chunk_result_df = await anyio.to_thread.run_sync(
                            partial(
                                download_multiple_sits,
                                gdf_chunk,
                                satellite,
                                reducers=reducers,
                                subsampling_max_pixels=subsampling_max_pixels,
                            )
                        )
                        chunk_result_df["chunk_id"] = chunk_id

                    whole_result_df = pd.concat([whole_result_df, chunk_result_df])
                except KeyboardInterrupt:
                    raise KeyboardInterrupt  # noqa: B904
                except Exception as e:
                    logger.error(f"download_multiple_sits_anyio_{chunk_id}_{satellite.shortName} = {e}")  # noqa: TRY400
                    indexes_with_errors.append(chunk_id)

        async with anyio.create_task_group() as tg:
            for cid in chunk_ids:
                tg.start_soon(download_task, cid)

    await run_download(all_chunk_ids, concurrency=initial_concurrency, timeout=initial_timeout)

    if indexes_with_errors:
        await run_download(indexes_with_errors, concurrency=retry_concurrency, timeout=retry_timeout)

    queue_listener.stop()

    whole_result_df = whole_result_df.sort_values(by=["indexnum"], kind="stable").reset_index(drop=True)

    return whole_result_df


def download_multiple_sits_chunks_multithread(
    gdf: gpd.GeoDataFrame,
    satellite: AbstractSatellite,
    reducers: list[str] | None = None,
    subsampling_max_pixels: float = 1_000,
    chunksize: int = 10000,
    mini_chunksize: int = 10,
    initial_concurrency: int = 30,
    retry_concurrency: int = 10,
    force_redownload: bool = False,
) -> gpd.GeoDataFrame:
    if len(gdf) == 0:
        print("Empty GeoDataFrame, nothing to download")
        return pd.DataFrame()

    schema = pa.DataFrameSchema({
        "geometry": pa.Column("geometry", nullable=False),
        "start_date": pa.Column(
            pa.DateTime,
            nullable=False,
            checks=pa.Check.in_range(min_value=satellite.startDate, max_value=satellite.endDate),
        ),
        "end_date": pa.Column(
            pa.DateTime,
            nullable=False,
            checks=pa.Check.in_range(min_value=satellite.startDate, max_value=satellite.endDate),
        ),
    })
    schema.validate(gdf, lazy=True)

    gdf = gdf.copy()
    add_indexnum_column(gdf)

    hashlib_gdf = create_gdf_hash(gdf)
    output_path = pathlib.Path("data/temp") / f"{satellite.shortName}_{hashlib_gdf}_{chunksize}"
    output_path.mkdir(parents=True, exist_ok=True)
    existing_chunks = {int(f.stem) for f in output_path.glob("*.parquet") if f.stem.isdigit()}

    gdf = quadtree_clustering(gdf, max_size=1000)

    num_chunks = (len(gdf) + chunksize - 1) // chunksize

    with tqdm(total=len(gdf), desc="Downloading multiple sits", smoothing=0.5) as pbar:
        for current_chunk in range(num_chunks):
            if (not force_redownload) and (current_chunk in existing_chunks):
                continue

            chunk_df = download_multiple_sits_multithread(
                gdf.iloc[current_chunk * chunksize : (current_chunk + 1) * chunksize],
                satellite,
                reducers=reducers,
                subsampling_max_pixels=subsampling_max_pixels,
                mini_chunksize=mini_chunksize,
                num_threads_rush=initial_concurrency,
                num_threads_retry=retry_concurrency,
                pbar=pbar,
            )
            chunk_df["timestamp"] = pd.to_datetime(chunk_df.timestamp)
            for col in chunk_df.columns:
                if col not in ["timestamp", "indexnum"]:
                    chunk_df[col] = chunk_df[col].astype(np.float16)

            chunk_df.to_parquet(f"{output_path / str(current_chunk)}.parquet")

    whole_result_df = pd.DataFrame()
    for f in sorted(output_path.glob("*.parquet"), key=lambda p: int(p.stem)):
        df = pd.read_parquet(f)
        whole_result_df = pd.concat([whole_result_df, df], ignore_index=True)

    whole_result_df = whole_result_df.sort_values(by=["indexnum"], kind="stable").reset_index(drop=True)

    return whole_result_df


def download_multiple_sits_task_gdrive(
    gdf: gpd.GeoDataFrame,
    satellite: AbstractSatellite,
    file_stem: str,
    reducers: list[str] | None = None,
    subsampling_max_pixels: float = 1_000,
    taskname: str = "",
    gee_save_folder: str = "GEE_EXPORTS",
) -> ee.batch.Task:
    if taskname == "":
        taskname = file_stem

    add_indexnum_column(gdf)
    fc = ee_gdf_to_feature_collection(gdf)
    ee_expression = ee.FeatureCollection(
        fc.map(
            partial(
                satellite.compute,
                reducers=reducers,
                subsampling_max_pixels=subsampling_max_pixels,
            )
        )
    ).flatten()

    task = ee.batch.Export.table.toDrive(
        collection=ee_expression,
        description=taskname,
        fileFormat="CSV",
        fileNamePrefix=file_stem,
        folder=gee_save_folder,
        selectors=["00_indexnum", "01_timestamp", *satellite.selectedBands],
    )

    return task


def download_multiple_sits_task_gcs(
    gdf: gpd.GeoDataFrame,
    satellite: AbstractSatellite,
    bucket_name: str,
    file_path: str,
    reducers: list[str] | None = None,
    subsampling_max_pixels: float = 1_000,
    taskname: str = "",
) -> ee.batch.Task:
    if taskname == "":
        taskname = file_path

    add_indexnum_column(gdf)
    fc = ee_gdf_to_feature_collection(gdf)
    ee_expression = ee.FeatureCollection(
        fc.map(
            partial(
                satellite.compute,
                reducers=reducers,
                subsampling_max_pixels=subsampling_max_pixels,
            )
        )
    ).flatten()

    task = ee.batch.Export.table.toCloudStorage(
        bucket=bucket_name,
        collection=ee_expression,
        description=taskname,
        fileFormat="CSV",
        fileNamePrefix=file_path,
        selectors=["00_indexnum", "01_timestamp", *satellite.selectedBands],
    )

    return task


def download_multiple_sits_chunks_gdrive(
    gdf: gpd.GeoDataFrame,
    satellite: AbstractSatellite,
    reducers: list[str] | None = None,
    subsampling_max_pixels: float = 1_000,
    cluster_size: int = 500,
    gee_save_folder: str = "GEE_EXPORTS",
    force_redownload: bool = False,
    wait: bool = True,
) -> None:
    if len(gdf) == 0:
        print("Empty GeoDataFrame, nothing to download")
        return None

    schema = pa.DataFrameSchema({
        "geometry": pa.Column("geometry", nullable=False),
        "start_date": pa.Column(
            pa.DateTime,
            nullable=False,
            checks=pa.Check.in_range(min_value=satellite.startDate, max_value=satellite.endDate),
        ),
        "end_date": pa.Column(
            pa.DateTime,
            nullable=False,
            checks=pa.Check.in_range(min_value=satellite.startDate, max_value=satellite.endDate),
        ),
    })
    schema.validate(gdf, lazy=True)

    task_mgr = GEETaskManager()  # To handle the new tasks

    tasks_df = ee_get_tasks_status()
    tasks_df = tasks_df[tasks_df.description.str.startswith("agl_multiple_sits_")]
    completed_or_running_tasks = set(
        tasks_df.description.apply(lambda x: x.split("_", 1)[0] + "_" + x.split("_", 2)[2]).tolist()
    )  # The task is the same, no matter who started it

    add_indexnum_column(gdf)
    gdf = quadtree_clustering(gdf, cluster_size)
    username = getpass.getuser().replace("_", "")
    hashname = create_gdf_hash(gdf)

    for cluster_id in tqdm(sorted(gdf.cluster_id.unique())):
        cluster_id = int(cluster_id)

        if (force_redownload) or (
            f"agl_multiple_sits_{satellite.shortName}_{hashname}_{cluster_id}" not in completed_or_running_tasks
        ):
            task = download_multiple_sits_task_gdrive(
                gdf[gdf.cluster_id == cluster_id],
                satellite,
                f"{satellite.shortName}_{hashname}_{cluster_id}",
                reducers=reducers,
                subsampling_max_pixels=subsampling_max_pixels,
                taskname=f"agl_{username}_multiple_sits_{satellite.shortName}_{hashname}_{cluster_id}",
                gee_save_folder=gee_save_folder,
            )

            task_mgr.add(task)

    task_mgr.start()  # Start all tasks at once allows user to cancel them before submitted to GEE

    if wait:
        task_mgr.wait()


def download_multiple_sits_chunks_gcs(
    gdf: gpd.GeoDataFrame,
    satellite: AbstractSatellite,
    bucket_name: str,
    reducers: list[str] | None = None,
    subsampling_max_pixels: float = 1_000,
    cluster_size: int = 500,
    force_redownload: bool = False,
    wait: bool = True,
) -> None | pd.DataFrame:
    if len(gdf) == 0:
        print("Empty GeoDataFrame, nothing to download")
        return None

    schema = pa.DataFrameSchema({
        "geometry": pa.Column("geometry", nullable=False),
        "start_date": pa.Column(
            pa.DateTime,
            nullable=False,
            checks=pa.Check.in_range(min_value=satellite.startDate, max_value=satellite.endDate),
        ),
        "end_date": pa.Column(
            pa.DateTime,
            nullable=False,
            checks=pa.Check.in_range(min_value=satellite.startDate, max_value=satellite.endDate),
        ),
    })
    schema.validate(gdf, lazy=True)

    task_mgr = GEETaskManager()
    tasks_df = ee_get_tasks_status()
    tasks_df = tasks_df[tasks_df.description.str.startswith("agl_")].reset_index(drop=True)
    completed_or_running_tasks = set(
        tasks_df.description.apply(lambda x: x.split("_", 1)[0] + "_" + x.split("_", 2)[2]).tolist()
    )  # The task is the same, no matter who started it

    add_indexnum_column(gdf)
    gdf = quadtree_clustering(gdf, cluster_size)
    username = getpass.getuser().replace("_", "")
    hashname = create_gdf_hash(gdf)

    gcs_save_folder = f"agl/{satellite.shortName}_{hashname}"
    metadata_dict: dict[str, str] = {}
    metadata_dict |= log_dict_function_call_summary(["gdf", "satellite"])
    metadata_dict |= satellite.log_dict()
    metadata_dict["user"] = username
    metadata_dict["creation_date"] = pd.Timestamp.now().strftime("%Y-%m-%d %H:%M:%S")

    with open(f"gs://{bucket_name}/{gcs_save_folder}/metadata.json", "w") as f:
        json.dump(metadata_dict, f, indent=4)

    with open(f"gs://{bucket_name}/{gcs_save_folder}/geodataframe.parquet", "wb") as f:
        gdf.to_parquet(f, compression="brotli")

    file_uris = []

    for cluster_id in tqdm(sorted(gdf.cluster_id.unique())):
        cluster_id = int(cluster_id)

        if (force_redownload) or (
            f"agl_multiple_sits_{satellite.shortName}_{hashname}_{cluster_id}" not in completed_or_running_tasks
        ):
            # TODO: Also skip if the file already exists in GCS
            task = download_multiple_sits_task_gcs(
                gdf[gdf.cluster_id == cluster_id],
                satellite,
                reducers=reducers,
                subsampling_max_pixels=subsampling_max_pixels,
                bucket_name=bucket_name,
                file_path=f"{gcs_save_folder}/{cluster_id}",
                taskname=f"agl_{username}_multiple_sits_{satellite.shortName}_{hashname}_{cluster_id}",
            )

            task_mgr.add(task)

        file_uris.append(f"gs://{bucket_name}/{gcs_save_folder}/{satellite.shortName}_{hashname}/{cluster_id}")

    task_mgr.start()

    if wait:
        task_mgr.wait()

        df = pd.DataFrame()
        for file_uri in file_uris:
            with open(file_uri, "rb") as f:
                sub_df = pd.read_csv(f)
                df = pd.concat([df, sub_df], ignore_index=True)

        remove_underscore_in_df(df)
        return df
    else:
        return None



from functools import partial

import ee

from agrigee_lite.ee_utils import (
    ee_add_indexes_to_image,
    ee_filter_img_collection_invalid_pixels,
    ee_get_number_of_pixels,
    ee_get_reducers,
    ee_map_bands_and_doy,
    ee_safe_remove_borders,
)
from agrigee_lite.sat.abstract_satellite import RadarSatellite


class Sentinel1(RadarSatellite):
    """
    Satellite abstraction for Sentinel-1.

    Sentinel-1 is a constellation of two Earth observation satellites,
    operated by ESA, designed for radar imaging of land and ocean surfaces. Sentinel-1B suffered a power failure in December 2021 and is no longer operational.

    Parameters
    ----------
    polarization : list of str, optional
        List of polarizations to select. Defaults to 'VV' and 'VH' for land applications.
    orbit_pass : str, optional
        Preferred orbit pass ('ASCENDING' or 'DESCENDING'). Defaults to ASCENDING.

    Satellite Information
    ---------------------
    +-----------------------------------+------------------------+
    | Field                             | Value                  |
    +-----------------------------------+------------------------+
    | Name                              | Sentinel-1             |
    | Revisit Time                      | 6 days (constellation) |
    | Revisit Time (after 2021)          | 12 days (only S1A)    |
    | Pixel Size                        | ~10 meters             |
    | Coverage                          | Global                 |
    +-----------------------------------+------------------------+

    Collection Dates
    ----------------
    +------------------+------------+----------------+
    | Collection Type  | Start Date | End Date        |
    +------------------+------------+----------------+
    | GRD (Ground Range Detected) | 2014-10-03 | present   |
    +------------------+------------+----------------+

    Band Information
    ------------
    +------------+---------------+--------------+----------------------+
    | Band Name  | Original Band | Resolution   | Spectral Wavelength   |
    +------------+---------------+--------------+----------------------+
    | VV         | C-band         | ~10 m        | 5.405 GHz (~5.5 cm)   |
    | VH         | C-band         | ~10 m        | 5.405 GHz (~5.5 cm)   |
    +------------+---------------+--------------+----------------------+
    """

    def __init__(
        self,
        bands: list[str] | None = None,
        indices: list[str] | None = None,
        ascending: bool = True,
    ):
        if bands is None:
            bands = ["vv", "vh"]

        if indices is None:
            indices = []

        super().__init__()

        self.ascending: bool = ascending
        self.imageCollectionName: str = "COPERNICUS/S1_GRD"
        self.pixelSize: int = 10

        # full mission start (S-1A launch)
        self.startDate: str = "2014-10-03"
        self.endDate: str = "2050-01-01"
        self.shortName: str = "s1a" if ascending else "s1d"

        # original → product band
        self.availableBands: dict[str, str] = {"vv": "VV", "vh": "VH"}

        self.selectedBands: list[tuple[str, str, str]] = [
            (band, f"{(n + 10):02}_{band}") for n, band in enumerate(bands)
        ]

        self.selectedIndices: list[str] = [
            (self.availableIndices[indice_name], indice_name, f"{(n + 40):02}_{indice_name}")
            for n, indice_name in enumerate(indices)
        ]

    @staticmethod
    def _mask_edge(img: ee.Image) -> ee.Image:
        """
        Remove extreme low-backscatter areas (edges / layover)

        Parameters
        ----------
        img : ee.Image
            Unfiltered Sentinel-1 image

        Returns
        -------
        ee.Image
            Filtered Sentinel-1 image
        """

        edge = img.lt(-30.0)
        valid = img.mask().And(edge.Not())
        return img.updateMask(valid)

    def imageCollection(self, ee_feature: ee.Feature) -> ee.ImageCollection:
        ee_geometry = ee_feature.geometry()
        ee_start = ee_feature.get("s")
        ee_end = ee_feature.get("e")

        ee_filter = ee.Filter.And(ee.Filter.bounds(ee_geometry), ee.Filter.date(ee_start, ee_end))

        s1_img = (
            ee.ImageCollection(self.imageCollectionName)
            .filter(ee_filter)
            .filter(
                ee.Filter.And(
                    ee.Filter.listContains("transmitterReceiverPolarisation", "VV"),
                    ee.Filter.listContains("transmitterReceiverPolarisation", "VH"),
                )
            )
            .filter(ee.Filter.eq("orbitProperties_pass", "ASCENDING" if self.ascending else "DESCENDING"))
            .map(self._mask_edge)
            .select(list(self.availableBands.values()), list(self.availableBands.keys()))
        )

        if self.selectedIndices:
            s1_img = s1_img.map(
                partial(ee_add_indexes_to_image, indexes=[expression for (expression, _, _) in self.selectedIndices])
            )

        s1_img = s1_img.select(
            [natural_band_name for natural_band_name, _ in self.selectedBands]
            + [indice_name for _, indice_name, _ in self.selectedIndices],
            [numeral_band_name for _, numeral_band_name in self.selectedBands]
            + [numeral_indice_name for _, _, numeral_indice_name in self.selectedIndices],
        )

        s1_img = ee_filter_img_collection_invalid_pixels(s1_img, ee_geometry, self.pixelSize, 20)

        return ee.ImageCollection(s1_img)

    def compute(
        self,
        ee_feature: ee.Feature,
        subsampling_max_pixels: float,
        reducers: list[str] | None = None,
    ) -> ee.FeatureCollection:
        ee_geometry = ee_feature.geometry()
        ee_geometry = ee_safe_remove_borders(ee_geometry, self.pixelSize, 35000)
        ee_feature = ee_feature.setGeometry(ee_geometry)

        s1_img = self.imageCollection(ee_feature)

        features = s1_img.map(
            partial(
                ee_map_bands_and_doy,
                ee_feature=ee_feature,
                pixel_size=self.pixelSize,
                subsampling_max_pixels=ee_get_number_of_pixels(ee_geometry, subsampling_max_pixels, self.pixelSize),
                reducer=ee_get_reducers(reducers),
            )
        )

        return features



from functools import partial

import anyio
import geopandas as gpd
import pandas as pd
import pytest

import agrigee_lite as agl
from agrigee_lite.sat.abstract_satellite import AbstractSatellite
from tests.utils import assert_df_equivalence, get_all_satellites_for_test

all_satellites = get_all_satellites_for_test()
all_reducers = ["min", "max", "mean", "median", "std", "var", "p2", "p98", "kurt", "skew"]
all_date_types = ["doy", "year", "fyear"]


@pytest.mark.parametrize("satellite", all_satellites)
def test_satellites_in_single_sits(satellite: AbstractSatellite) -> None:
    gdf = gpd.read_parquet("tests/data/gdf.parquet")
    row = gdf.iloc[0]

    sits = agl.get.sits(row.geometry, row.start_date, row.end_date, satellite)
    original_sits = pd.read_parquet(f"tests/data/sits/0_{satellite.shortName}.parquet")
    assert_df_equivalence(sits, original_sits)


@pytest.mark.parametrize("satellite", all_satellites)
@pytest.mark.parametrize("reducer", all_reducers)
def test_reducers_of_all_satellites_in_single_sits(satellite: AbstractSatellite, reducer: str) -> None:
    gdf = gpd.read_parquet("tests/data/gdf.parquet")
    row = gdf.iloc[0]

    sits = agl.get.sits(row.geometry, row.start_date, row.end_date, satellite, reducers=[reducer])
    original_sits = pd.read_parquet(f"tests/data/sits/0_{satellite.shortName}_{reducer}.parquet")
    assert_df_equivalence(sits, original_sits)


def test_multiple_reducers() -> None:
    gdf = gpd.read_parquet("tests/data/gdf.parquet")
    satellite = agl.sat.Sentinel2(bands=["swir1", "nir"])
    row = gdf.iloc[0]

    sits = agl.get.sits(
        row.geometry, row.start_date, row.end_date, satellite, ["mode", "p95", "p5", "var"], ["doy"], 0.3
    )
    original_sits = pd.read_parquet("tests/data/sits/multiple_reducers.parquet")

    assert_df_equivalence(sits, original_sits)


def test_download_multiple_sits_multithread() -> None:
    gdf = gpd.read_parquet("tests/data/gdf.parquet")
    satellite = agl.sat.Sentinel2(bands=["swir1", "nir"])
    sits = agl.get.multiple_sits(gdf.iloc[0:2], satellite, ["skew", "p13"], ["doy"], 0.3)
    original_sits = pd.read_parquet("tests/data/sits/multithread.parquet")

    assert_df_equivalence(sits, original_sits)


@pytest.mark.parametrize("date_type", all_date_types)
def test_date_type(date_type: str) -> None:
    gdf = gpd.read_parquet("tests/data/gdf.parquet")
    satellite = agl.sat.Sentinel2(bands=["nir", "green"])
    row = gdf.iloc[0]

    sits = agl.get.sits(row.geometry, row.start_date, row.end_date, satellite, ["kurt", "mode"], [date_type], 100)
    original_sits = pd.read_parquet(f"tests/data/sits/datetype_{date_type}.parquet")

    assert_df_equivalence(sits, original_sits)


def test_all_date_types() -> None:
    gdf = gpd.read_parquet("tests/data/gdf.parquet")
    satellite = agl.sat.Sentinel2(bands=["swir1", "swir2", "re4"])
    row = gdf.iloc[0]

    sits = agl.get.sits(row.geometry, row.start_date, row.end_date, satellite, ["kurt", "mode"], all_date_types, 200)
    original_sits = pd.read_parquet("tests/data/sits/all_date_types.parquet")

    assert_df_equivalence(sits, original_sits)


# def test_download_multiple_sits_async() -> None:
#     from agrigee_lite.get.sits import __download_multiple_sits_async

#     gdf = gpd.read_parquet("tests/data/gdf.parquet")
#     satellite = agl.sat.Sentinel2(bands=["swir1", "nir"])
#     sits = anyio.run(
#         partial(__download_multiple_sits_async, gdf.iloc[0:2].copy(), satellite, ["skew", "p13"], ["doy"], 1),
#         backend_options={"use_uvloop": True},
#     )
#     original_sits = pd.read_parquet("tests/data/sits/async.parquet")

#     assert_df_equivalence(sits, original_sits)



from functools import partial

import ee

from agrigee_lite.ee_utils import (
    ee_add_indexes_to_image,
    ee_filter_img_collection_invalid_pixels,
    ee_get_number_of_pixels,
    ee_get_reducers,
    ee_map_bands_and_doy,
    ee_safe_remove_borders,
)
from agrigee_lite.sat.abstract_satellite import OpticalSatellite


class Modis(OpticalSatellite):
    def __init__(self, bands: list[str] | None = None, indices: list[str] | None = None) -> None:
        if bands is None:
            bands = ["red", "nir"]

        if indices is None:
            indices = []

        super().__init__()

        self.shortName = "modis"
        self.pixelSize = 250
        self.startDate = "2000-02-24"
        self.endDate = "2050-01-01"

        self._terra_vis = "MODIS/061/MOD09GQ"
        self._terra_qa = "MODIS/061/MOD09GA"
        self._aqua_vis = "MODIS/061/MYD09GQ"
        self._aqua_qa = "MODIS/061/MYD09GA"

        self.availableBands = {
            "red": "sur_refl_b01",
            "nir": "sur_refl_b02",
        }

        self.selectedBands: list[tuple[str, str, str]] = [
            (band, f"{(n + 10):02}_{band}") for n, band in enumerate(bands)
        ]

        self.selectedIndices: list[str] = [
            (self.availableIndices[indice_name], indice_name, f"{(n + 40):02}_{indice_name}")
            for n, indice_name in enumerate(indices)
        ]

    @staticmethod
    def _mask_modis_clouds(img: ee.Image) -> ee.Image:
        """Bit-test bit 10 of *state_1km* (value 0 = clear)."""
        qa = img.select("state_1km")
        bit_mask = 1 << 10
        return img.updateMask(qa.bitwiseAnd(bit_mask).eq(0))

    def imageCollection(self, ee_feature: ee.Feature) -> ee.ImageCollection:
        """
        Build the merged, cloud-masked Terra + Aqua collection *exactly*
        like the stand-alone helper did.
        """
        ee_geometry = ee_feature.geometry()
        ee_filter = ee.Filter.And(
            ee.Filter.bounds(ee_geometry),
            ee.Filter.date(ee_feature.get("s"), ee_feature.get("e")),
        )

        def _base(vis: str, qa: str) -> ee.ImageCollection:
            return (
                ee.ImageCollection(vis)
                .linkCollection(ee.ImageCollection(qa), ["state_1km"])
                .filter(ee_filter)
                .map(self._mask_modis_clouds)
                .select(
                    list(self.availableBands.values()),
                    list(self.availableBands.keys()),
                )
            )

        terra = _base(self._terra_vis, self._terra_qa)
        aqua = _base(self._aqua_vis, self._aqua_qa)

        modis_imgc = terra.merge(aqua)

        modis_imgc = modis_imgc.map(
            lambda img: ee.Image(img).addBands(ee.Image(img).add(100).divide(16_100), overwrite=True)
        )

        if self.selectedIndices:
            modis_imgc = modis_imgc.map(
                partial(ee_add_indexes_to_image, indexes=[expression for (expression, _, _) in self.selectedIndices])
            )

        modis_imgc = modis_imgc.select(
            [natural_band_name for natural_band_name, _ in self.selectedBands]
            + [indice_name for _, indice_name, _ in self.selectedIndices],
            [numeral_band_name for _, numeral_band_name in self.selectedBands]
            + [numeral_indice_name for _, _, numeral_indice_name in self.selectedIndices],
        )

        modis_imgc = ee_filter_img_collection_invalid_pixels(modis_imgc, ee_geometry, self.pixelSize, 2)

        return ee.ImageCollection(modis_imgc)

    def compute(
        self,
        ee_feature: ee.Feature,
        subsampling_max_pixels: float,
        reducers: list[str] | None = None,
    ) -> ee.FeatureCollection:
        """Sample time series of median reflectance within *ee_feature*."""
        geom = ee_feature.geometry()
        geom = ee_safe_remove_borders(geom, self.pixelSize // 2, 190_000)
        ee_feature = ee_feature.setGeometry(geom)

        modis = self.imageCollection(ee_feature)

        feats = modis.map(
            partial(
                ee_map_bands_and_doy,
                ee_feature=ee_feature,
                pixel_size=self.pixelSize,
                subsampling_max_pixels=ee_get_number_of_pixels(geom, subsampling_max_pixels, self.pixelSize),
                reducer=ee_get_reducers(reducers),
            )
        )
        return feats



from warnings import simplefilter

import pandas as pd

from agrigee_lite import get, sat, vis
from agrigee_lite.ee_utils import ee_get_tasks_status as get_all_tasks
from agrigee_lite.misc import quadtree_clustering, random_points_from_gdf

simplefilter(action="ignore", category=pd.errors.PerformanceWarning)
pd.options.mode.copy_on_write = True



import ee

from agrigee_lite.ee_utils import ee_safe_remove_borders
from agrigee_lite.sat.abstract_satellite import DataSourceSatellite


class MapBiomas(DataSourceSatellite):
    def __init__(self) -> None:
        super().__init__()
        self.imageAsset: str = (
            "projects/mapbiomas-public/assets/brazil/lulc/collection9/mapbiomas_collection90_integration_v1"
        )
        self.pixelSize: int = 30
        self.startDate = "1985-02-24"
        self.endDate = "2023-12-31"
        self.shortName = "mapbiomasmajclass"

    def compute(
        self,
        ee_feature: ee.Feature,
        subsampling_max_pixels: float | None = None,
        reducers: list[str] | None = None,
    ) -> ee.FeatureCollection:
        ee_geometry = ee_feature.geometry()
        ee_geometry = ee_safe_remove_borders(ee_geometry, self.pixelSize, 50000)
        ee_feature = ee_feature.setGeometry(ee_geometry)

        mb_image = ee.Image(self.imageAsset)

        ee_start = ee.Feature(ee_feature).get("s")
        ee_end = ee.Feature(ee_feature).get("e")
        start_year = ee.Date(ee_start).get("year")
        end_year = ee.Date(ee_end).get("year")
        indexnum = ee.Feature(ee_feature).get("0")

        years = ee.List.sequence(start_year, end_year)

        def _feat_for_year(year: ee.Number) -> ee.Feature:
            year_num = ee.Number(year).toInt()
            year_str = year_num.format()
            band_in = ee.String("classification_").cat(year_str)
            img = mb_image.select([band_in], [year_str])

            mode_dict = img.reduceRegion(
                reducer=ee.Reducer.mode(),
                geometry=ee_geometry,
                scale=self.pixelSize,
                maxPixels=1e13,
                bestEffort=True,
            )
            clazz = ee.Number(mode_dict.get(year_str)).round()

            percent = (
                img.eq(clazz)
                .reduceRegion(
                    reducer=ee.Reducer.mean(),
                    geometry=ee_geometry,
                    scale=self.pixelSize,
                    maxPixels=1e13,
                    bestEffort=True,
                )
                .get(year_str)
            )
            return ee.Feature(
                None,
                {
                    "00_indexnum": indexnum,
                    "02_year": year,
                    "10_class": clazz,
                    "11_percent": percent,
                },
            )

        features = years.map(_feat_for_year)
        return ee.FeatureCollection(features)

    def __str__(self) -> str:
        return self.shortName

    def __repr__(self) -> str:
        return self.shortName



import json
import os
import random
import string

import ee
import geopandas as gpd
import numpy as np
import pandas as pd


def ee_map_bands_and_doy(
    ee_img: ee.Image,
    ee_feature: ee.Feature,
    pixel_size: int,
    subsampling_max_pixels: ee.Number,
    reducer: ee.Reducer,
) -> ee.Feature:
    ee_img = ee.Image(ee_img)
    ee_stats = ee_img.reduceRegion(
        reducer=reducer,
        geometry=ee_feature.geometry(),
        scale=pixel_size,
        maxPixels=subsampling_max_pixels,
        bestEffort=True,
    ).map(lambda _, value: ee.Number(ee.Algorithms.If(ee.Algorithms.IsEqual(value, None), 0, value)))

    ee_stats = ee_stats.set("01_timestamp", ee.Date(ee_img.date()).format("YYYY-MM-dd"))
    ee_stats = ee_stats.set("00_indexnum", ee_feature.get("0"))

    return ee.Feature(None, ee_stats)


def ee_map_valid_pixels(img: ee.Image, ee_geometry: ee.Geometry, pixel_size: int) -> ee.Image:
    mask = ee.Image(img).select([0]).gt(0)

    valid_pixels = ee.Number(
        mask.rename("valid")
        .reduceRegion(
            reducer=ee.Reducer.count(),
            geometry=ee_geometry,
            scale=pixel_size,
            maxPixels=1e13,
            bestEffort=True,
        )
        .get("valid")
    )

    return ee.Image(img.set("ZZ_USER_VALID_PIXELS", valid_pixels))


def ee_cloud_probability_mask(img: ee.Image, threshold: float, invert: bool = False) -> ee.Image:
    mask = img.select(["cloud"]).gte(threshold) if invert else img.select(["cloud"]).lt(threshold)

    return img.updateMask(mask).select(img.bandNames().remove("cloud"))


def ee_gdf_to_feature_collection(gdf: gpd.GeoDataFrame) -> ee.FeatureCollection:
    gdf = gdf[["00_indexnum", "geometry", "start_date", "end_date"]]

    gdf["start_date"] = gdf["start_date"].dt.strftime("%Y-%m-%d")
    gdf["end_date"] = gdf["end_date"].dt.strftime("%Y-%m-%d")

    gdf.rename(
        columns={"start_date": "s", "end_date": "e", "00_indexnum": "0"}, inplace=True
    )  # saving memory when uploading geojson to GEE

    geo_json = os.path.join(os.getcwd(), "".join(random.choice(string.ascii_lowercase) for i in range(6)) + ".geojson")  # noqa: S311
    gdf = gdf.to_crs(4326)
    gdf.to_file(geo_json, driver="GeoJSON")

    with open(os.path.abspath(geo_json), encoding="utf-8") as f:
        json_dict = json.load(f)

    if json_dict["type"] == "FeatureCollection":
        for feature in json_dict["features"]:
            if feature["geometry"]["type"] != "Point":
                feature["geometry"]["geodesic"] = True
        features = ee.FeatureCollection(json_dict)

    os.remove(geo_json)

    return features


def ee_img_to_numpy(ee_img: ee.Image, ee_geometry: ee.Geometry, scale: int) -> np.ndarray:
    ee_img = ee.Image(ee_img)
    ee_geometry = ee.Geometry(ee_geometry).bounds()

    projection = ee.Projection("EPSG:4326").atScale(scale).getInfo()
    chip_size = round(ee_geometry.perimeter(0.1).getInfo() / (4 * scale))  # type: ignore  # noqa: PGH003

    scale_y = -projection["transform"][0]  # type: ignore  # noqa: PGH003
    scale_x = projection["transform"][4]  # type: ignore  # noqa: PGH003

    list_of_coordinates = ee.Array.cat(ee_geometry.coordinates(), 1).getInfo()

    x_min = list_of_coordinates[0][0]  # type: ignore  # noqa: PGH003
    y_max = list_of_coordinates[2][1]  # type: ignore  # noqa: PGH003
    coordinates = [x_min, y_max]

    chip_size = 1 if chip_size == 0 else chip_size

    img_in_bytes = ee.data.computePixels({
        "expression": ee_img,
        "fileFormat": "NUMPY_NDARRAY",
        "grid": {
            "dimensions": {"width": chip_size, "height": chip_size},
            "affineTransform": {
                "scaleX": scale_x,
                "scaleY": scale_y,
                "translateX": coordinates[0],
                "translateY": coordinates[1],
            },
            "crsCode": projection["crs"],  # type: ignore  # noqa: PGH003
        },
    })

    img_in_array = np.array(img_in_bytes.tolist()).astype(np.float32)
    img_in_array[np.isinf(img_in_array)] = 0
    img_in_array[np.isnan(img_in_array)] = 0

    return img_in_array


def ee_get_tasks_status() -> pd.DataFrame:
    tasks = ee.data.listOperations()

    if tasks:
        records = []
        for op in tasks:
            metadata = op.get("metadata", {})

            record = {
                "attempt": metadata.get("attempt"),
                "create_time": metadata.get("createTime"),
                "description": metadata.get("description"),
                "destination_uris": metadata.get("destinationUris", [None])[0],
                "done": op.get("done"),
                "end_time": metadata.get("endTime"),
                "name": op.get("name"),
                "priority": metadata.get("priority"),
                "progress": metadata.get("progress"),
                "script_uri": metadata.get("scriptUri"),
                "start_time": metadata.get("startTime"),
                "state": metadata.get("state"),
                "total_batch_eecu_usage_seconds": metadata.get("batchEecuUsageSeconds", 0.0),
                "type": metadata.get("type"),
                "update_time": metadata.get("updateTime"),
            }
            records.append(record)

        df = pd.DataFrame(records)
        df["create_time"] = pd.to_datetime(df.create_time, format="mixed")
        df["end_time"] = pd.to_datetime(df.end_time, format="mixed")
        df["start_time"] = pd.to_datetime(df.start_time, format="mixed")
        df["update_time"] = pd.to_datetime(df.update_time, format="mixed")

        df["estimated_cost_usd_tier_1"] = (df.total_batch_eecu_usage_seconds / (60 * 60)) * 0.40
        df["estimated_cost_usd_tier_2"] = (df.total_batch_eecu_usage_seconds / (60 * 60)) * 0.28
        df["estimated_cost_usd_tier_3"] = (df.total_batch_eecu_usage_seconds / (60 * 60)) * 0.16

    else:  # If no tasks are found, create an empty DataFrame with the same columns
        df = pd.DataFrame(
            columns=[
                "attempt",
                "create_time",
                "description",
                "destination_uris",
                "done",
                "end_time",
                "name",
                "priority",
                "progress",
                "script_uri",
                "start_time",
                "state",
                "total_batch_eecu_usage_seconds",
                "type",
                "update_time",
            ]
        )

    return df


def ee_get_reducers(reducer_names: list[str] | None = None) -> ee.Reducer:  # noqa: C901
    if reducer_names is None:
        reducer_names = ["median"]

    names = [n.lower() for n in reducer_names]

    pct_vals = sorted({int(n[1:]) for n in names if n.startswith("p")})

    reducers = []
    for n in names:
        if n == "min":
            reducers.append(ee.Reducer.min())
        elif n == "max":
            reducers.append(ee.Reducer.max())
        elif n == "mean":
            reducers.append(ee.Reducer.mean())
        elif n == "median":
            reducers.append(ee.Reducer.median())
        elif n == "kurt":
            reducers.append(ee.Reducer.kurtosis())
        elif n == "skew":
            reducers.append(ee.Reducer.skew())
        elif n == "std":
            reducers.append(ee.Reducer.stdDev())
        elif n == "var":
            reducers.append(ee.Reducer.variance())
        elif n == "mode":
            reducers.append(ee.Reducer.mode())
        elif n.startswith("p"):
            continue
        else:
            raise ValueError(f"Unknown reducer: '{n}'")  # noqa: TRY003

    if pct_vals:
        reducers.append(ee.Reducer.percentile(pct_vals))

    reducer = reducers[0]
    for r in reducers[1:]:
        reducer = reducer.combine(r, None, True)

    return reducer


def ee_filter_img_collection_invalid_pixels(
    ee_img_collection: ee.ImageCollection, ee_geometry: ee.Geometry, pixel_size: int, min_valid_pixels: int = 20
) -> ee.ImageCollection:
    min_valid_pixels = ee.Algorithms.If(
        ee_geometry.area(),
        ee.Number(min_valid_pixels),
        ee.Number(1),
    )

    ee_img_collection = ee_img_collection.map(lambda i: ee_map_valid_pixels(i, ee_geometry, pixel_size)).filter(
        ee.Filter.gte("ZZ_USER_VALID_PIXELS", min_valid_pixels)
    )

    ee_img_collection = (
        ee_img_collection.map(lambda img: img.set("ZZ_USER_TIME_DUMMY", img.date().format("YYYY-MM-dd")))
        .sort("ZZ_USER_TIME_DUMMY")
        .distinct("ZZ_USER_TIME_DUMMY")
    )

    return ee_img_collection


def ee_get_number_of_pixels(ee_geometry: ee.Geometry, subsampling_max_pixels: float, pixel_size: int) -> ee.Number:
    # -- maxPixels logic (absolute or fraction of footprint) -- #
    if subsampling_max_pixels > 1:
        return ee.Number(subsampling_max_pixels)
    else:
        pixel_area = ee.Number(pixel_size).pow(2)
        total_pixels = ee_geometry.area().divide(pixel_area)
        return total_pixels.multiply(subsampling_max_pixels).toInt()


def ee_safe_remove_borders(ee_geometry: ee.Geometry, border_size: int, area_lower_bound: int) -> ee.Geometry:
    return ee.Geometry(
        ee.Algorithms.If(
            ee_geometry.buffer(-border_size).area().gte(area_lower_bound),
            ee_geometry.buffer(-border_size),
            ee_geometry,
        )
    )


def ee_add_indexes_to_image(image: ee.Image, indexes: list[str]) -> ee.Image:
    for index in indexes:
        calculated = image.expression(index, {"i": image})
        image = image.addBands(calculated, None, True)

    return image



import re

VEGETATION_INDICES_LIST = [
    "evi = 2.5 * ((i.nir - i.red)/(i.nir + 6 * i.red - 7.5 * i.blue + 1))",
    "evi2 = 2.5 * ((i.nir - i.red)/(i.nir + 2.4 * i.red + 1))",
    "ndvi = (i.nir - i.red)/(i.nir + i.red)",
    "ndwi = (i.nir - i.swir1)/(i.nir + i.swir1)",
    "mndwi = (i.green - i.swir1)/(i.green + i.swir1)",
    "vhvv = (i.vv - i.vh)/(i.vv + i.vh)",
]

VEGETATION_INDICES = {}

for item in VEGETATION_INDICES_LIST:
    key, expression = item.split("=", 1)
    key = key.strip()
    expression = expression.strip()
    bands = set(re.findall(r"i\.([a-z0-9_]+)", expression))
    VEGETATION_INDICES[key] = {"expression": item, "required_bands": bands}



import os

import ee
import geopandas as gpd
import numpy as np

import agrigee_lite as agl
from agrigee_lite.sat.abstract_satellite import AbstractSatellite
from tests.utils import get_all_date_types_for_test, get_all_reducers_for_test, get_all_satellites_for_test


def download_img_for_test(satellite: AbstractSatellite) -> None:
    gdf = gpd.read_parquet("tests/data/gdf.parquet")
    row = gdf.iloc[0]

    imgs = agl.get.images(row.geometry, row.start_date, row.end_date, satellite)
    np.savez_compressed(f"tests/data/imgs/0_{satellite.shortName}.npz", data=imgs)


def download_sits_for_test(satellite: AbstractSatellite) -> None:
    gdf = gpd.read_parquet("tests/data/gdf.parquet")
    row = gdf.iloc[0]

    sits = agl.get.sits(row.geometry, row.start_date, row.end_date, satellite)
    sits.to_parquet(f"tests/data/sits/0_{satellite.shortName}.parquet")


def download_sits_for_test_with_reducers(satellite: AbstractSatellite) -> None:
    gdf = gpd.read_parquet("tests/data/gdf.parquet")
    row = gdf.iloc[0]

    all_reducers = get_all_reducers_for_test()

    for reducer in all_reducers:
        sits = agl.get.sits(row.geometry, row.start_date, row.end_date, satellite, reducers=[reducer])
        sits.to_parquet(f"tests/data/sits/0_{satellite.shortName}_{reducer}.parquet")


def download_for_test_download_multiple_sits() -> None:
    gdf = gpd.read_parquet("tests/data/gdf.parquet")
    satellite = agl.sat.Sentinel2(bands=["red"])
    sits = agl.get.multiple_sits(gdf.iloc[0:2], satellite, ["kurt", "median"], ["doy"], 0.7)
    sits.to_parquet("tests/data/sits/multithread.parquet")


# def download_for_test_download_multiple_sits_async() -> None:
#     from agrigee_lite.get.sits import __download_multiple_sits_async

#     gdf = gpd.read_parquet("tests/data/gdf.parquet")
#     satellite = agl.sat.Sentinel2(bands=["swir1", "nir"])
#     sits = anyio.run(
#         partial(__download_multiple_sits_async, gdf.iloc[0:2], satellite, ["skew", "p13"], ["doy"], 1),
#         backend_options={"use_uvloop": True},
#     )
#     sits.to_parquet("tests/data/sits/async.parquet")


def download_for_test_multiple_reducers() -> None:
    gdf = gpd.read_parquet("tests/data/gdf.parquet")
    satellite = agl.sat.Sentinel2(bands=["swir1", "nir"])
    row = gdf.iloc[0]

    sits = agl.get.sits(
        row.geometry, row.start_date, row.end_date, satellite, ["mode", "p95", "p5", "var"], ["doy"], 0.3
    )
    sits.to_parquet("tests/data/sits/multiple_reducers.parquet")


def download_for_test_date_type(date_type: str) -> None:
    gdf = gpd.read_parquet("tests/data/gdf.parquet")
    satellite = agl.sat.Sentinel2(bands=["nir", "green"])
    row = gdf.iloc[0]

    sits = agl.get.sits(row.geometry, row.start_date, row.end_date, satellite, ["kurt", "mode"], [date_type], 100)
    sits.to_parquet(f"tests/data/sits/datetype_{date_type}.parquet")


def test_all_date_types(all_date_types: list[str]) -> None:
    gdf = gpd.read_parquet("tests/data/gdf.parquet")
    satellite = agl.sat.Sentinel2(bands=["swir1", "swir2", "re4"])
    row = gdf.iloc[0]

    sits = agl.get.sits(row.geometry, row.start_date, row.end_date, satellite, ["kurt", "mode"], all_date_types, 200)
    sits.to_parquet("tests/data/sits/all_date_types.parquet")


if __name__ == "__main__":
    ee.Initialize(opt_url="https://earthengine-highvolume.googleapis.com", project="ee-paulagibrim")
    all_satellites = get_all_satellites_for_test()
    all_date_types = get_all_date_types_for_test()

    os.makedirs("tests/data/imgs", exist_ok=True)
    os.makedirs("tests/data/sits", exist_ok=True)

    for date_type in all_date_types:
        print("Downloading date type", date_type, "...")
        download_for_test_date_type(date_type)

    test_all_date_types(all_date_types)
    download_for_test_download_multiple_sits()
    # download_for_test_download_multiple_sits_async()
    download_for_test_multiple_reducers()

    for satellite in all_satellites:
        print("Downloading satellite", satellite.shortName, "...")
        download_img_for_test(satellite)
        download_sits_for_test(satellite)
        download_sits_for_test_with_reducers(satellite)



from agrigee_lite.get.image import download_multiple_images as images
from agrigee_lite.get.sits import download_multiple_sits_chunks_gcs as multiple_sits_gcs
from agrigee_lite.get.sits import download_multiple_sits_chunks_gdrive as multiple_sits_gdrive
from agrigee_lite.get.sits import download_multiple_sits_chunks_multithread as multiple_sits
from agrigee_lite.get.sits import download_single_sits as sits



import geopandas as gpd
import matplotlib.pyplot as plt
import pandas as pd
from shapely import Polygon

from agrigee_lite.get.sits import download_multiple_sits_chunks_multithread, download_single_sits
from agrigee_lite.misc import compute_index_from_df
from agrigee_lite.numpy_indices import ALL_NUMPY_INDICES
from agrigee_lite.sat.abstract_satellite import AbstractSatellite


def year_fraction(dt: pd.Series) -> pd.Series:
    year = dt.year
    start_of_year = pd.Timestamp(year=year, month=1, day=1)
    end_of_year = pd.Timestamp(year=year + 1, month=1, day=1)
    fraction = (dt - start_of_year) / (end_of_year - start_of_year)
    return year + fraction


def visualize_single_sits(
    geometry: Polygon,
    start_date: pd.Timestamp | str,
    end_date: pd.Timestamp | str,
    satellite: AbstractSatellite,
    band_or_indice_to_plot: str,
    reducer: str = "median",
    ax: plt.Axes | None = None,
    color: str = "blue",
    alpha: float = 1,
) -> None:
    long_sits = download_single_sits(geometry, start_date, end_date, satellite, reducers=[reducer])

    if band_or_indice_to_plot in ALL_NUMPY_INDICES:
        y = compute_index_from_df(long_sits, ALL_NUMPY_INDICES[band_or_indice_to_plot])
    else:
        y = long_sits[band_or_indice_to_plot].values

    long_sits["timestamp"] = pd.to_datetime(long_sits["timestamp"])

    if ax is None:
        plt.plot(
            long_sits.timestamp,
            y,
            color=color,
            alpha=alpha,
        )
        plt.scatter(
            long_sits.timestamp,
            y,
            color=color,
        )
    else:
        ax.plot(long_sits.timestamp, y, color=color, alpha=alpha, label=satellite.shortName)
        ax.scatter(
            long_sits.timestamp,
            y,
            color=color,
        )


def visualize_multiple_sits(
    gdf: gpd.GeoDataFrame,
    satellite: AbstractSatellite,
    band_or_indice_to_plot: str,
    reducer: str = "median",
    ax: plt.Axes | None = None,
    color: str = "blue",
    alpha: float = 0.5,
) -> None:
    long_sits = download_multiple_sits_chunks_multithread(gdf, satellite, reducers=[reducer])

    if band_or_indice_to_plot in ALL_NUMPY_INDICES:
        long_sits["y"] = compute_index_from_df(long_sits, ALL_NUMPY_INDICES[band_or_indice_to_plot])

    for indexnumm in long_sits.indexnum.unique():
        indexnumm_df = long_sits[long_sits.indexnum == indexnumm].reset_index(drop=True).copy()
        indexnumm_df["timestamp"] = indexnumm_df.timestamp.apply(year_fraction)
        indexnumm_df["timestamp"] = indexnumm_df["timestamp"] - indexnumm_df["timestamp"].min().round()

        y = (
            indexnumm_df["y"]
            if band_or_indice_to_plot in ALL_NUMPY_INDICES
            else indexnumm_df[band_or_indice_to_plot].values
        )

        if ax is None:
            plt.plot(
                indexnumm_df.timestamp,
                y,
                color=color,
                alpha=alpha,
            )
        else:
            ax.plot(indexnumm_df.timestamp, y, color=color, alpha=alpha, label=satellite.shortName)



import geopandas as gpd
import numpy as np
import pytest

import agrigee_lite as agl
from agrigee_lite.sat.abstract_satellite import AbstractSatellite
from tests.utils import assert_np_array_equivalence, get_all_satellites_for_test

all_satellites = get_all_satellites_for_test()


@pytest.mark.parametrize("satellite", all_satellites)
def test_download_images(satellite: AbstractSatellite) -> None:
    gdf = gpd.read_parquet("tests/data/gdf.parquet")
    row = gdf.iloc[0]

    imgs = agl.get.images(row.geometry, row.start_date, row.end_date, satellite)
    original_imgs = np.load(f"tests/data/imgs/0_{satellite.shortName}.npz")
    assert_np_array_equivalence(imgs, original_imgs["data"], 0)



from functools import partial

import ee

from agrigee_lite.ee_utils import (
    ee_add_indexes_to_image,
    ee_filter_img_collection_invalid_pixels,
    ee_get_number_of_pixels,
    ee_get_reducers,
    ee_map_bands_and_doy,
    ee_safe_remove_borders,
)
from agrigee_lite.sat.abstract_satellite import OpticalSatellite


def remove_l_toa_tough_clouds(img: ee.Image) -> ee.Image:
    img = ee.Image(img)
    img = ee.Algorithms.Landsat.simpleCloudScore(img)

    mask = img.select(["cloud"]).lte(15)
    img = img.updateMask(mask)
    return img.select(img.bandNames().remove("cloud"))


def ee_l_mask(img: ee.Image) -> ee.Image:
    qa = img.select("cloudq")
    mask = (
        qa.bitwiseAnd(1 << 3)
        .And(qa.bitwiseAnd(1 << 8).Or(qa.bitwiseAnd(1 << 9)))
        .Or(qa.bitwiseAnd(1 << 1))
        .Or(qa.bitwiseAnd(1 << 4).And(qa.bitwiseAnd(1 << 10).Or(qa.bitwiseAnd(1 << 11))))
        .Or(qa.bitwiseAnd(1 << 5))
        .Or(qa.bitwiseAnd(1 << 7))
        .Or(qa.bitwiseAnd(1 << 2))
    )

    return img.updateMask(mask.Not()).select(img.bandNames().remove("cloudq"))


def ee_l_apply_sr_scale_factors(img: ee.Image) -> ee.Image:
    img = ee.Image(img)
    optical_bands = img.select("SR_B.").multiply(0.0000275).add(-0.2)
    # thermal_bands = img.select("ST_B6").multiply(0.00341802).add(149.0)
    return img.addBands(optical_bands, None, True)  # .addBands(thermal_bands, None, True)


class AbstractLandsat(OpticalSatellite):
    _DEFAULT_BANDS: list[str] = [  # noqa: RUF012
        "blue",
        "green",
        "red",
        "nir",
        "swir1",
        "swir2",
    ]

    def __init__(
        self,
        *,
        sensor_code: str,  # e.g. "LT05"
        toa_band_map: dict[str, str],
        sr_band_map: dict[str, str],
        short_base: str,  # e.g. "l5"
        start_date: str,  # sensor-specific
        end_date: str,  # sensor-specific
        bands: list[str] | None = None,
        indices: list[str] | None = None,
        use_sr: bool = True,
        tier: int = 1,
    ) -> None:
        super().__init__()

        if indices is None:
            indices = []

        bands = bands or self._DEFAULT_BANDS
        self.useSr = use_sr
        self.tier = tier
        self.pixelSize: int = 30

        self.startDate: str = start_date
        self.endDate: str = end_date

        suffix = "L2" if use_sr else "TOA"
        self.imageCollectionName = f"LANDSAT/{sensor_code}/C02/T{tier}_{suffix}"
        self.shortName: str = f"{short_base}sr" if use_sr else short_base

        self.availableBands = sr_band_map if use_sr else toa_band_map
        self.availableBands["cloudq"] = "QA_PIXEL"

        self.selectedBands: list[tuple[str, str, str]] = [
            (band, f"{(n + 10):02}_{band}") for n, band in enumerate(bands)
        ]

        self.selectedIndices: list[str] = [
            (self.availableIndices[indice_name], indice_name, f"{(n + 40):02}_{indice_name}")
            for n, indice_name in enumerate(indices)
        ]

    def imageCollection(self, ee_feature: ee.Feature) -> ee.ImageCollection:
        geom = ee_feature.geometry()
        ee_filter = ee.Filter.And(
            ee.Filter.bounds(geom),
            ee.Filter.date(ee_feature.get("s"), ee_feature.get("e")),
        )

        col = ee.ImageCollection(self.imageCollectionName).filter(ee_filter)
        col = col.map(ee_l_apply_sr_scale_factors) if self.useSr else col.map(remove_l_toa_tough_clouds)

        col = col.select(list(self.availableBands.values()), list(self.availableBands.keys()))
        col = col.map(ee_l_mask)

        if self.selectedIndices:
            col = col.map(
                partial(ee_add_indexes_to_image, indexes=[expression for (expression, _, _) in self.selectedIndices])
            )

        col = col.select(
            [natural_band_name for natural_band_name, _ in self.selectedBands]
            + [indice_name for _, indice_name, _ in self.selectedIndices],
            [numeral_band_name for _, numeral_band_name in self.selectedBands]
            + [numeral_indice_name for _, _, numeral_indice_name in self.selectedIndices],
        )

        col = ee_filter_img_collection_invalid_pixels(col, geom, self.pixelSize, 12)
        return ee.ImageCollection(col)

    def compute(
        self,
        ee_feature: ee.Feature,
        subsampling_max_pixels: float,
        reducers: list[str] | None = None,
    ) -> ee.FeatureCollection:
        geom = ee_feature.geometry()
        geom = ee_safe_remove_borders(geom, self.pixelSize, 50000)
        ee_feature = ee_feature.setGeometry(geom)

        col = self.imageCollection(ee_feature)
        features = col.map(
            partial(
                ee_map_bands_and_doy,
                ee_feature=ee_feature,
                pixel_size=self.pixelSize,
                subsampling_max_pixels=ee_get_number_of_pixels(geom, subsampling_max_pixels, self.pixelSize),
                reducer=ee_get_reducers(reducers),
            )
        )
        return features


class Landsat5(AbstractLandsat):
    def __init__(
        self,
        bands: list[str] | None = None,
        indices: list[str] | None = None,
        use_sr: bool = True,
        tier: int = 1,
    ):
        toa = {"blue": "B1", "green": "B2", "red": "B3", "nir": "B4", "swir1": "B5", "swir2": "B7"}
        sr = {
            "blue": "SR_B1",
            "green": "SR_B2",
            "red": "SR_B3",
            "nir": "SR_B4",
            "swir1": "SR_B5",
            "swir2": "SR_B7",
        }
        super().__init__(
            indices=indices,
            sensor_code="LT05",
            toa_band_map=toa,
            sr_band_map=sr,
            short_base="l5",
            start_date="1984-03-01",
            end_date="2013-05-05",
            bands=bands,
            use_sr=use_sr,
            tier=tier,
        )


class Landsat7(AbstractLandsat):
    def __init__(
        self,
        bands: list[str] | None = None,
        indices: list[str] | None = None,
        use_sr: bool = True,
        tier: int = 1,
    ):
        toa = {"blue": "B1", "green": "B2", "red": "B3", "nir": "B4", "swir1": "B5", "swir2": "B7"}
        sr = {
            "blue": "SR_B1",
            "green": "SR_B2",
            "red": "SR_B3",
            "nir": "SR_B4",
            "swir1": "SR_B5",
            "swir2": "SR_B7",
        }
        super().__init__(
            indices=indices,
            sensor_code="LE07",
            toa_band_map=toa,
            sr_band_map=sr,
            short_base="l7",
            start_date="1999-04-15",
            end_date="2022-04-06",
            bands=bands,
            use_sr=use_sr,
            tier=tier,
        )


class Landsat8(AbstractLandsat):
    def __init__(
        self,
        bands: list[str] | None = None,
        indices: list[str] | None = None,
        use_sr: bool = True,
        tier: int = 1,
    ):
        toa = {"blue": "B2", "green": "B3", "red": "B4", "nir": "B5", "swir1": "B6", "swir2": "B7"}
        sr = {
            "blue": "SR_B2",
            "green": "SR_B3",
            "red": "SR_B4",
            "nir": "SR_B5",
            "swir1": "SR_B6",
            "swir2": "SR_B7",
        }
        super().__init__(
            indices=indices,
            sensor_code="LC08",
            toa_band_map=toa,
            sr_band_map=sr,
            short_base="l8",
            start_date="2013-04-11",
            end_date="2050-01-01",
            bands=bands,
            use_sr=use_sr,
            tier=tier,
        )


class Landsat9(AbstractLandsat):
    def __init__(
        self,
        bands: list[str] | None = None,
        indices: list[str] | None = None,
        use_sr: bool = True,
        tier: int = 1,
    ):
        toa = {"blue": "B2", "green": "B3", "red": "B4", "nir": "B5", "swir1": "B6", "swir2": "B7"}
        sr = {
            "blue": "SR_B2",
            "green": "SR_B3",
            "red": "SR_B4",
            "nir": "SR_B5",
            "swir1": "SR_B6",
            "swir2": "SR_B7",
        }
        super().__init__(
            indices=indices,
            sensor_code="LC09",
            toa_band_map=toa,
            sr_band_map=sr,
            short_base="l9",
            start_date="2021-11-01",
            end_date="2050-01-01",
            bands=bands,
            use_sr=use_sr,
            tier=tier,
        )



from agrigee_lite.sat.landsat import Landsat5, Landsat7, Landsat8, Landsat9
from agrigee_lite.sat.mapbiomas import MapBiomas
from agrigee_lite.sat.modis import Modis
from agrigee_lite.sat.sentinel1 import Sentinel1
from agrigee_lite.sat.sentinel2 import Sentinel2



import concurrent.futures
import hashlib
import inspect
import warnings
from collections import deque
from collections.abc import Callable
from functools import lru_cache, wraps
from typing import ParamSpec, TypeVar

import geopandas as gpd
import numpy as np
import pandas as pd
from shapely.geometry import MultiPolygon, Point, Polygon
from topojson import Topology
from tqdm.std import tqdm


def build_quadtree_iterative(gdf: gpd.GeoDataFrame, max_size: int = 1000) -> list[int]:
    queue: deque[tuple[gpd.GeoDataFrame, int]] = deque()
    queue.append((gdf, 0))
    leaves = []

    while queue:
        subset, depth = queue.popleft()
        n = len(subset)
        if n <= max_size:
            leaves.append(subset.index.to_numpy())
            continue

        dim = "centroid_x" if depth % 2 == 0 else "centroid_y"

        subset_sorted = subset.sort_values(by=dim)
        median_idx = n // 2
        median_val = subset_sorted.iloc[median_idx][dim]

        left = subset_sorted[subset_sorted[dim] <= median_val]
        right = subset_sorted[subset_sorted[dim] > median_val]

        queue.append((left, depth + 1))
        queue.append((right, depth + 1))

    return leaves


def build_quadtree(gdf: gpd.GeoDataFrame, max_size: int = 1000, depth: int = 0) -> list[int]:
    n = len(gdf)
    if n <= max_size:
        return [gdf.index.to_numpy()]

    dim = "centroid_x" if depth % 2 == 0 else "centroid_y"

    gdf_sorted = gdf.sort_values(by=dim)

    median_idx = n // 2
    median_val = gdf_sorted.iloc[median_idx][dim]

    left = gdf_sorted[gdf_sorted[dim] <= median_val]
    right = gdf_sorted[gdf_sorted[dim] > median_val]

    left_clusters = build_quadtree(left, max_size, depth + 1)
    right_clusters = build_quadtree(right, max_size, depth + 1)

    return left_clusters + right_clusters


def simplify_gdf(gdf: gpd.GeoDataFrame, tol: float = 0.001) -> gpd.GeoDataFrame:
    """
    1. Detect duplicate geometries once, using WKB-hex as a stable key.
    2. Run TopoJSON simplification only on the unique geometries.
    3. Propagate the simplified result back to every original row.
    """
    gdf = gdf.copy()

    # ------------------------------------------------------------------
    # 1.  Build a geometry-only frame and keep just the unique geometries
    # ------------------------------------------------------------------
    gdf["_geom_key"] = gdf.geometry.apply(lambda g: g.wkb_hex)  # fast, deterministic
    unique_gdf = gdf[["_geom_key", "geometry"]].drop_duplicates("_geom_key")

    # ---------------------------------------------------------------
    # 2.  Simplify the unique geometries once with Topology.toposimplify
    # ---------------------------------------------------------------
    topo = Topology(unique_gdf[["geometry"]], prequantize=False)
    topo = topo.toposimplify(tol, prevent_oversimplify=True)
    simplified_unique = topo.to_gdf()

    # topo.to_gdf() returns rows in the same order, so align keys back
    simplified_unique["_geom_key"] = unique_gdf["_geom_key"].values

    # -------------------------------------------------------
    # 3.  Merge the simplified geometries back to the original
    # -------------------------------------------------------
    out = (
        gdf.drop(columns="geometry")
        .merge(simplified_unique[["_geom_key", "geometry"]], on="_geom_key", how="left")
        .drop(columns="_geom_key")
        .set_geometry("geometry")
    )
    out.index = gdf.index  # keep the original ordering
    return out


def _simplify_cluster(cluster: gpd.GeoDataFrame, cluster_id: int) -> tuple[int, gpd.GeoSeries]:
    simplified = simplify_gdf(cluster)
    return cluster_id, simplified.geometry


def quadtree_clustering(
    gdf: gpd.GeoDataFrame,
    max_size: int = 1_000,
) -> gpd.GeoDataFrame:
    gdf = gdf.copy()

    # Centroid columns (ignore CRS warning)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=UserWarning)
        gdf["centroid_x"] = gdf.geometry.centroid.x
        gdf["centroid_y"] = gdf.geometry.centroid.y

    # Build quadtree and label clusters
    clusters = build_quadtree_iterative(gdf, max_size=max_size)

    cluster_array = np.zeros(len(gdf), dtype=int)
    for i, cluster_indexes in enumerate(clusters):
        cluster_array[cluster_indexes] = i

    gdf["cluster_id"] = cluster_array
    gdf = gdf.sort_values(by=["cluster_id", "centroid_x"]).reset_index(drop=True)

    unique_cluster_ids = gdf["cluster_id"].unique()

    with concurrent.futures.ProcessPoolExecutor() as executor:
        futures = {
            executor.submit(_simplify_cluster, gdf[gdf.cluster_id == cluster_id][["geometry"]], cluster_id): cluster_id
            for cluster_id in unique_cluster_ids
        }

        for future in tqdm(
            concurrent.futures.as_completed(futures),
            total=len(futures),
            desc="Simplifying clusters",
            smoothing=0.5,
        ):
            cluster_id, simplified_geom = future.result()
            gdf.loc[gdf["cluster_id"] == cluster_id, "geometry"] = simplified_geom.values

    return gdf


def create_gdf_hash(gdf: gpd.GeoDataFrame) -> str:
    gdf_copy = gdf[["geometry", "start_date", "end_date"]].copy()

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        gdf_copy["centroid_x"] = gdf_copy.geometry.centroid.x
        gdf_copy["centroid_y"] = gdf_copy.geometry.centroid.y

    gdf_copy = gdf_copy.drop(columns=["geometry"])

    hash_values = pd.util.hash_pandas_object(gdf_copy).values
    return hashlib.sha1(hash_values).hexdigest()  # type: ignore  # noqa: PGH003, S324


P = ParamSpec("P")
R = TypeVar("R")


def cached(func: Callable[P, R]) -> Callable[P, R]:
    cached_func = lru_cache()(func)

    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        return cached_func(*args, **kwargs)  # type: ignore  # noqa: PGH003

    return wrapper


def remove_underscore_in_df(df: pd.DataFrame | gpd.GeoDataFrame) -> None:
    df.columns = [column.split("_", 1)[1] for column in df.columns.tolist()]


def compute_index_from_df(df: pd.DataFrame, np_function: Callable) -> np.ndarray:
    sig = inspect.signature(np_function)
    kwargs = {}

    index_name = str(np_function.__name__).split("np_")[1]

    if index_name in df.columns.tolist():
        return df[index_name].to_numpy()

    for param_name, param in sig.parameters.items():
        if param_name in df.columns:
            kwargs[param_name] = df[param_name].values
        else:
            if param.default is not inspect._empty:
                kwargs[param_name] = param.default
            else:
                raise ValueError(  # noqa: TRY003
                    f"DataFrame is missing a column '{param_name}', "
                    f"required by {np_function.__name__}, and there's no default."
                )

    return np_function(**kwargs)


def add_indexnum_column(df: pd.DataFrame) -> None:
    if "00_indexnum" not in df.columns:
        if not (df.index.to_numpy() == np.arange(len(df))).all():
            raise ValueError(  # noqa: TRY003
                "The index must be sequential from 0 to N-1. To do this, use gdf.reset_index(drop=True) before executing this function."
            )
        df["00_indexnum"] = range(len(df))


def log_dict_function_call_summary(ignore: list[str] | None = None) -> dict[str, dict[str, str]]:
    frame = inspect.currentframe().f_back
    func_name = frame.f_code.co_name
    args, _, _, values = inspect.getargvalues(frame)
    ignore = ignore or []
    args_dict = {str(arg): str(values[arg]) for arg in args if arg not in ignore}
    return {func_name: args_dict}


def create_grid_centroids_numpy(geometry: Polygon | MultiPolygon, n_cells=10) -> np.ndarray:
    try:
        xmin, ymin, xmax, ymax = geometry.bounds
        cell_size = (xmax - xmin) / n_cells

        num_cols = int(np.ceil((xmax - xmin) / cell_size))
        num_rows = int(np.ceil((ymax - ymin) / cell_size))
        max_points = num_cols * num_rows

        centroids = np.empty((max_points, 2), dtype=np.float32)
        count = 0

        for x in np.arange(xmin + cell_size / 2, xmax, cell_size):
            for y in np.arange(ymin + cell_size / 2, ymax, cell_size):
                point = Point(x, y)
                if geometry.contains(point):
                    centroids[count] = [x, y]
                    count += 1

        if count >= n_cells:
            return centroids[np.random.choice(count, size=n_cells, replace=False)]
        if count == n_cells:
            return centroids[np.random.choice(count, size=n_cells, replace=True)]
        else:  # count < n_cells:
            return np.zeros((n_cells, 2), dtype=np.float32)
    except:  # noqa: E722
        return np.zeros((n_cells, 2), dtype=np.float32)


def generate_grid_random_points_from_gdf(gdf: gpd.GeoDataFrame, num_points_per_geometry=10) -> gpd.GeoDataFrame:
    centroids = np.empty((num_points_per_geometry * gdf.geometry.nunique(), 2), dtype=np.float32)
    geometry_ids = np.empty((num_points_per_geometry * gdf.geometry.nunique()), dtype=np.int32)

    for n, (_, row) in enumerate(
        tqdm(gdf[["geometry", "geometry_id"]].drop_duplicates().iterrows(), total=gdf.geometry.nunique())
    ):
        geom = row.geometry
        geometry_id = row.geometry_id
        centroids_sub = create_grid_centroids_numpy(geom, n_cells=num_points_per_geometry)
        centroids[n * num_points_per_geometry : (n + 1) * num_points_per_geometry, :] = centroids_sub
        geometry_ids[n * num_points_per_geometry : (n + 1) * num_points_per_geometry] = geometry_id

    gdf = gpd.GeoDataFrame(geometry=[Point(lon, lat) for lon, lat in centroids], crs=gdf.crs)
    gdf["geometry_id"] = geometry_ids

    return gdf


def random_points_from_gdf(
    gdf: gpd.GeoDataFrame, num_points_per_geometry: int = 10, buffer: int = -10
) -> gpd.GeoDataFrame:
    if buffer != 0:
        gdf = gdf.copy()
        gdf = quadtree_clustering(gdf)
        gdf["geometry"] = gdf.to_crs(gdf.estimate_utm_crs()).buffer(-10).to_crs("EPSG:4326")

    gdf["geometry_id"] = pd.factorize(gdf["geometry"])[0]
    points_gdf = generate_grid_random_points_from_gdf(gdf, num_points_per_geometry)
    points_gdf = points_gdf.merge(
        gdf.drop(columns=["geometry"]).reset_index().rename(columns={"index": "original_index"}),
        on="geometry_id",
        how="inner",
    )

    return points_gdf
